/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAB_DEBUG			FALSE
#define FE_SAB_COPY_DEBUG		FALSE

#include "surface.pmh"

namespace fe
{

SurfaceAccessibleBase::SurfaceAccessibleBase(void):
	m_paging(FALSE)
{
}

SurfaceAccessibleBase::~SurfaceAccessibleBase(void)
{
}

void SurfaceAccessibleBase::reset(void)
{
	if(m_spScope.isValid())
	{
		SurfaceModel surfaceModelRV;
		surfaceModelRV.bind(m_spScope);
		m_weakRecord=surfaceModelRV.createRecord();

		FEASSERT(m_weakRecord.isValid());

		m_spRecordGroup=new RecordGroup();
		m_spRecordGroup->add(m_weakRecord);
	}
	else
	{
		//* TODO check this
		FEASSERT(FALSE);

		m_weakRecord=Record();
		m_spRecordGroup=NULL;
	}
}

Protectable* SurfaceAccessibleBase::clone(Protectable* pInstance)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::clone\n");
#endif

	SurfaceAccessibleBase* pSurfaceAccessibleBase= pInstance?
			dynamic_cast<SurfaceAccessibleBase*>(pInstance):
			new SurfaceAccessibleBase();

	//* HACK save and reload to duplicate RecordGroup

	FEASSERT(m_spScope.isValid());
	const String tempname="temp.rg";
	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,tempname);

	sp<RecordGroup>& rspRecordGroup=pSurfaceAccessibleBase->m_spRecordGroup;
	rspRecordGroup=RecordView::loadRecordGroup(m_spScope,tempname);

	//* update surface with new Record
	RecordGroup::iterator it=rspRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	Record record=spRA->getRecord(0);
	pSurfaceAccessibleBase->bind(record);

	return pSurfaceAccessibleBase;
}

BWORD SurfaceAccessibleBase::load(String a_filename)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::load\n");
#endif

	m_spRecordGroup=RecordView::loadRecordGroup(m_spScope,a_filename);
	if(m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	RecordGroup::iterator it=m_spRecordGroup->begin();
	sp<RecordArray> spRA= *it;
	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(spRA);

	m_weakRecord=spRA->getRecord(0);

	Record record=m_weakRecord;
	Instance model;
	model.set<Record>(registry()->master()->typeMaster(),record);
	bind(model);

	return TRUE;
}

BWORD SurfaceAccessibleBase::save(String a_filename)
{
#if	FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::save\n");
#endif

	FEASSERT(m_spScope.isValid());
	FEASSERT(m_spRecordGroup.isValid());

	if(m_spScope.isNull() || m_spRecordGroup.isNull())
	{
		return FALSE;
	}

	RecordView::saveRecordGroup(m_spScope,m_spRecordGroup,a_filename);
	return TRUE;
}

void SurfaceAccessibleBase::groupNames(Array<String>& a_rNameArray,
	SurfaceAccessibleI::Element a_element) const
{
	a_rNameArray.clear();

	//* TODO
}

void SurfaceAccessibleBase::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleBase::attributeSpecs\n");

	a_rSpecs.clear();

	const String layoutName=elementLayout(a_element);
	sp<Layout> spElementLayout=m_spScope->lookupLayout(layoutName);
	sp<TypeMaster> spTypeMaster=registry()->master()->typeMaster();

//	feLog("SurfaceAccessibleBase::attributeSpecs layout %s %p\n",
//			layoutName.c_str(),spElementLayout.raw());

	if(spElementLayout.isValid())
	{
		const U32 attrCount=spElementLayout->attributeCount();

//		feLog("SurfaceAccessibleBase::attributeSpecs attrCount %d\n",
//				attrCount);

		for(U32 attrIndex=0;attrIndex<attrCount;attrIndex++)
		{
			sp<fe::Attribute> spAttribute=spElementLayout->attribute(attrIndex);
			const String attrName=spAttribute->name();

			if(attrName==":CT" || attrName==":SN")
			{
				continue;
			}

			const String attrTypeName=spTypeMaster->lookupBaseType(
					spAttribute->type());
			String typeName;

			if(attrTypeName=="string")
			{
				typeName="string";
			}
			else if(attrTypeName=="I32")
			{
				typeName="integer";
			}
			else if(attrTypeName=="F32")
			{
				typeName="real";
			}
			else if(attrTypeName=="vector3f")
			{
				typeName="vector3";
			}

//			feLog("SurfaceAccessibleBase::attributeSpecs %d/%d %s %s %s\n",
//					attrIndex,attrCount,
//					attrTypeName.c_str(),typeName.c_str(),attrName.c_str());

			SurfaceAccessibleI::Spec spec;
			spec.set(commonName(attrName),typeName);

			a_rSpecs.push_back(spec);
		}
	}
}

//* TODO fix for Maya, or provide Maya version with filtering
void SurfaceAccessibleBase::append(
	sp<SurfaceAccessibleI> a_spSurfaceAccessible,
	const SpatialTransform* a_pTransform,sp<FilterI> a_spFilter)
{
#if FE_SAB_COPY_DEBUG
	feLog("SurfaceAccessibleBase::append\n");
#endif

	sp<SurfaceAccessorI> spInputVertices=
			a_spSurfaceAccessible->accessor(e_primitive,e_vertices);

	sp<SurfaceAccessorI> spOutputVertices=accessor(e_primitive,e_vertices,
			SurfaceAccessibleI::e_createMissing);

	I32 offset[4];

	for(U32 elementPass=0;elementPass<4;elementPass++)
	{
		SurfaceAccessibleI::Element element;

		switch(elementPass)
		{
			case 0:
				element=e_point;
				break;
			case 1:
				element=e_vertex;
				break;
			case 2:
				element=e_primitive;
				break;
			case 3:
				element=e_detail;
				break;
			default:
				;
		}

		sp<SurfaceAccessorI> spOutputGeneric=accessor(element,e_generic);
		offset[elementPass]=
				spOutputGeneric.isValid()? spOutputGeneric->count(): 0;
	}

	if(spInputVertices.isValid() && spOutputVertices.isValid())
	{
		const I32 offsetIndex=offset[0];

		const U32 filterCount=a_spFilter.isValid()?
				a_spFilter->filterCount(): spInputVertices->count();

		I32 pointCount=0;

		//* point indices for each vertex per primitive
		Array< Array<I32> > primVerts;
		primVerts.resize(filterCount);

		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 primitiveIndex=a_spFilter.isValid()?
					a_spFilter->filter(filterIndex): filterIndex;

			Array<I32>& rVertList=primVerts[primitiveIndex];

			const U32 subCount=spInputVertices->subCount(primitiveIndex);

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spInputVertices->integer(primitiveIndex,subIndex);

				rVertList.push_back(offsetIndex+pointIndex);

				if(pointCount<pointIndex+1)
				{
					pointCount=pointIndex+1;
				}
			}
		}

		sp<SurfaceAccessorI> spPointPosition=accessor(e_point,e_position);

#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append output pointCount %d\n",
				pointCount);
#endif

		for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			spPointPosition->append();
		}

		//* WARNING currently, Maya ignores any point append and adds them here
		//* this is different than the Houdini version
		spOutputVertices->append(primVerts);
	}

	for(U32 elementPass=0;elementPass<4;elementPass++)
	{
		const I32 offsetIndex=offset[elementPass];

		SurfaceAccessibleI::Element element;

		switch(elementPass)
		{
			case 0:
				element=e_point;
				break;
			case 1:
				element=e_vertex;
				break;
			case 2:
				element=e_primitive;
				break;
			case 3:
				element=e_detail;
				break;
			default:
				;
		}

		Array<Spec> specs;
		a_spSurfaceAccessible->attributeSpecs(specs,element);

		const U32 elementCount=a_spSurfaceAccessible->count(element);
		const U32 specCount=specs.size();

#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append"
				" element %d elementCount %d offsetIndex %d specCount %d\n",
				element,elementCount,offsetIndex,specCount);
#endif

		for(U32 specIndex=0;specIndex<specCount;specIndex++)
		{
			const Spec& spec=specs[specIndex];
			const String& specName=spec.name();
			const String& specType=spec.typeName();

			sp<SurfaceAccessorI> spInputAccessor=
					a_spSurfaceAccessible->accessor(element,specName);

			sp<SurfaceAccessorI> spOutputAccessor=accessor(element,specName,
					SurfaceAccessibleI::e_createMissing);
			FEASSERT(spOutputAccessor.isValid());

			if(spInputAccessor.isNull() || spOutputAccessor.isNull())
			{
				feLog("SurfaceAccessibleBase::append %s %s accessor failure\n",
						specType.c_str(),specName.c_str());
				continue;
			}

#if FE_SAB_COPY_DEBUG
			feLog("SurfaceAccessibleBase::append %s %s count %d\n",
					specType.c_str(),specName.c_str(),elementCount);
#endif

			if(specType=="vector3")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					SpatialVector vector3=
							spInputAccessor->spatialVector(elementIndex);

					if(a_pTransform)
					{
						if(specName=="P")
						{
							transformVector(*a_pTransform,vector3,vector3);
						}
						else if(specName=="N")
						{
							rotateVector(*a_pTransform,vector3,vector3);
						}
						//* more?
					}

					spOutputAccessor->set(offsetIndex+elementIndex,vector3);
				}
			}
			else if(specType=="real")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->real(elementIndex));
				}
			}
			else if(specType=="integer")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->integer(elementIndex));
				}
			}
			else if(specType=="string")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->string(elementIndex));
				}
			}
		}
#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append output elementCount %d \n",
				count(element));
#endif
	}

	sp<SurfaceAccessorI> spInputProperties=
			a_spSurfaceAccessible->accessor(e_primitive,e_properties);

	sp<SurfaceAccessorI> spOutputProperties=accessor(e_primitive,e_properties,
			SurfaceAccessibleI::e_createMissing);

	if(spInputVertices.isValid() && spOutputVertices.isValid() &&
			spInputProperties.isValid() && spOutputProperties.isValid())
	{
		const I32 offsetIndex=offset[2];

		const U32 filterCount=a_spFilter.isValid()?
				a_spFilter->filterCount(): spInputVertices->count();

		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 primitiveIndex=a_spFilter.isValid()?
					a_spFilter->filter(filterIndex): filterIndex;

			const BWORD openCurve=spInputProperties->integer(primitiveIndex,
					SurfaceAccessibleI::e_openCurve);
			spOutputProperties->set(offsetIndex+filterIndex,
					SurfaceAccessibleI::e_openCurve,openCurve);
		}
	}
}

//* static
String SurfaceAccessibleBase::elementLayout(
	SurfaceAccessibleI::Element a_element)
{
	switch(a_element)
	{
		case e_point:
			return "pt";
		case e_vertex:
			return "vtx";
		case e_primitive:
			return "prim";
		case e_detail:
			return "detail";
		default:
			;
	}

	return "";
}

//* static
String SurfaceAccessibleBase::commonName(String a_attrName)
{
	if(a_attrName=="spc:at")
	{
		return "P";
	}
	else if(a_attrName=="spc:up")
	{
		return "N";
	}
	else if(a_attrName=="surf:uvw")
	{
		return "uv";
	}

	return a_attrName;
}

//* static
String SurfaceAccessibleBase::attributeName(String a_specName)
{
	if(a_specName=="P")
	{
		return "spc:at";
	}
	else if(a_specName=="N")
	{
		return "spc:up";
	}
	else if(a_specName=="uv")
	{
		return "surf:uvw";
	}

	return a_specName;
}

void SurfaceAccessibleBase::require(SurfaceAccessibleI::Element a_element,
	String a_attrName,String a_attrTypeName)
{
	const String layoutName=elementLayout(a_element);

	m_spScope->declare(layoutName);
	m_spScope->support(a_attrName,a_attrTypeName);
	m_spScope->populate(layoutName,a_attrName);
}

sp<SurfaceAccessorI> SurfaceAccessibleBase::accessor(
	Element a_element,String a_name,Creation a_create)
{
	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorBase> spAccessor(new SurfaceAccessorBase);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

		if(spAccessor->bind(a_element,a_name))
		{
			return spAccessor;
		}
	}

	return sp<SurfaceAccessorI>(NULL);
}

I32 SurfaceAccessibleBase::count(SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleBase::count element %d\n",a_element);

	switch(a_element)
	{
		case e_point:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleBase*>(this)->accessor(
					e_point,e_position);

			if(spAccessor.isNull())
			{
				feLog("SurfaceAccessibleBase::count point accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		case e_primitive:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleBase*>(this)->accessor(
					e_primitive,e_vertices);

			if(spAccessor.isNull())
			{
				feLog("SurfaceAccessibleBase::count primitive accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		default:
			;
	}

	return 0;
}

sp<SurfaceAccessorI> SurfaceAccessibleBase::accessor(
	Element a_element,Attribute a_attribute,Creation a_create)
{
	FEASSERT(m_weakRecord.isValid());

	sp<SurfaceAccessorBase> spAccessor(new SurfaceAccessorBase);

	if(spAccessor.isValid())
	{
		FEASSERT(m_weakRecord.isValid());

		spAccessor->bind(m_weakRecord);

		FEASSERT(m_weakRecord.isValid());

		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));

			FEASSERT(m_weakRecord.isValid());

		if(spAccessor->bind(a_element,a_attribute))
		{
			FEASSERT(m_weakRecord.isValid());

			return spAccessor;
		}

		feLog("SurfaceAccessibleBase::accessor"
				" failed element %d attr %d create %d\n",
				a_element,a_attribute,a_create);
	}

	FEASSERT(m_weakRecord.isValid());

	return sp<SurfaceAccessorI>(NULL);
}

BWORD SurfaceAccessibleBase::discard(SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute)
{
	String name;
	switch(a_attribute)
	{
		case SurfaceAccessibleI::e_position:
			name="P";
			break;
		case SurfaceAccessibleI::e_normal:
			name="N";
			break;
		case SurfaceAccessibleI::e_uv:
			name="uv";
			break;
		default:
			return FALSE;
	}

	return discard(a_element,name);
}

sp<SurfaceI> SurfaceAccessibleBase::surface(String a_group)
{
//	feLog("SurfaceAccessibleBase::surface(\"%s\")\n",a_group.c_str());

	//* TODO select form
	//* TODO group
	sp<SurfaceTrianglesAccessible> spSurface;

	if(registry().isValid())
	{
		spSurface=registry()->create("*.SurfaceTrianglesAccessible");
	}
	else
	{
		feLog("SurfaceAccessibleBase::surface registry is NULL\n");
	}
	if(spSurface.isValid())
	{
		spSurface->bind(sp<SurfaceAccessibleI>(this));
	}
	return spSurface;
}

sp<SurfaceI> SurfaceAccessibleBase::subSurface(U32 a_subIndex,String a_group)
{
	//* TODO

	return sp<SurfaceI>(NULL);
}

sp<SpannedRange> SurfaceAccessibleBase::atomize(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomize\n");
#endif

	if(a_atomicChange==e_pointsOfPrimitives)
	{
		const BWORD checkPages=TRUE;
		return atomizeConnectivity(a_atomicChange,a_group,
				a_desiredCount,checkPages);
	}

	//* fallback to non-atomic

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);
	const Element element=asPoints? e_point: e_primitive;
	sp<SurfaceAccessorI> spAccessor=accessor(element,e_generic);
	const I32 elementCount=spAccessor.isValid()? spAccessor->count(): 0;
	sp<SpannedRange> spSpannedRange(new SpannedRange());

	//* use everything

	if(a_group.empty())
	{
		if(elementCount>0)
		{
			spSpannedRange->nonAtomic().add(0,elementCount-1);
		}
		return spSpannedRange;
	}

	//* use group

//	feLog("SurfaceAccessibleBase::atomize group \"%s\"\n",a_group.c_str());

	I32 spanStart= -1;
	SpannedRange::MultiSpan& rMultiSpan=spSpannedRange->nonAtomic();

#if FALSE
	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const I32 elementPage=
				asPoints? pointPage(elementIndex): primitivePage(elementIndex);
		if(spanStart<0 && elementPage>=0)
		{
			//* begin span
			spanStart=elementIndex;
		}
		else if(spanStart>=0 && elementPage<0)
		{
			//* complete span
			rMultiSpan.add(spanStart,elementIndex-1);
			spanStart=-1;
		}
	}

	//* complete last span
	if(spanStart>=0)
	{
		rMultiSpan.add(spanStart,elementCount-1);

	}
#else
	const Element groupElement=asPoints? e_pointGroup: e_primitiveGroup;

	sp<SurfaceAccessorI> spElementGroup=accessor(groupElement,a_group);

	spElementGroup->appendGroupSpans(rMultiSpan);
#endif

	return spSpannedRange;
}

sp<SpannedRange> SurfaceAccessibleBase::atomizeConnectivity(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount,
	BWORD a_checkPages)
{
	if(a_checkPages)
	{
		preparePaging(a_atomicChange,a_group);
	}

	sp<SpannedRange> spSpannedRange(new SpannedRange());

	sp<SurfaceAccessorI> spPoints=accessor(e_point,e_position);
	sp<SurfaceAccessorI> spVertices=accessor(e_primitive,e_vertices);

	const U32 pointCount=spPoints->count();
	const U32 primitiveCount=spVertices->count();

	Array<U32> pointPatch(pointCount,0);
	Array<I32> primitivePatch(primitiveCount,0);

	Array<U32> pagePatch(primitiveCount+1,0);	//* pages <= prims+1
	Array<U32> histogram(primitiveCount+1,0);	//* patches <= prims+1

	U32 maxPatch=0;
	U32 maxPage=0;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		if(a_checkPages && primitivePage(primitiveIndex)<0)
		{
			FEASSERT(primitiveIndex<primitiveCount);
			primitivePatch[primitiveIndex]= -1;
			continue;
		}

		U32 lowPatch=0;
		U32 lowPage=0;
		BWORD multiPatch=FALSE;
		BWORD zeroPatches=FALSE;
		BWORD multiPage=FALSE;

		const U32 subCount=spVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=spVertices->integer(primitiveIndex,subIndex);
			FEASSERT(pointIndex<pointCount);
			U32& rPatch=pointPatch[pointIndex];
			FEASSERT(rPatch<=maxPatch);

			const I32 page=a_checkPages? pointPage(pointIndex): 0;
			if(page>0)
			{
				FEASSERT(page<=primitiveCount);
				const U32 existingPatch=pagePatch[page];
				FEASSERT(existingPatch<=maxPatch);

				if(existingPatch)
				{
					rPatch=existingPatch;
				}
				if(!lowPage)
				{
					lowPage=page;
				}
				else if(lowPage!=page)
				{
					multiPage=TRUE;
					if(lowPage>page)
					{
						lowPage=page;
					}
				}
			}
			if(rPatch)
			{
				if(!lowPatch)
				{
					lowPatch=rPatch;
				}
				else if(lowPatch!=rPatch)
				{
					multiPatch=TRUE;
					if(lowPatch>rPatch)
					{
						lowPatch=rPatch;
					}
				}
			}
			else
			{
				zeroPatches=TRUE;
			}
		}

		if(multiPage)
		{
			//* leave for postAtomic
			lowPatch=0;
		}
		else if(multiPatch)
		{
//			feLog("multiPatch\n");

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				const U32 patch=pointPatch[pointIndex];

//				feLog("prim %d sub %d/%d patch %d\n",
//						primitiveIndex,subIndex,subCount,patch);

				if(patch<1)
				{
//					feLog("assign patch %d to lowPatch %d\n",patch,lowPatch);

					pointPatch[pointIndex]=lowPatch;
				}
				else if(patch!=lowPatch)
				{
//					feLog("connected patch %d != lowPatch %d\n",patch,lowPatch);

					for(U32 pointOther=0;pointOther<pointCount;pointOther++)
					{
						FEASSERT(pointOther<pointCount);
						U32& rOtherPatch=pointPatch[pointOther];

						if(rOtherPatch==patch)
						{
//							feLog("%d merge patch %d into %d\n",
//									pointIndex,rOtherPatch,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPatch=lowPatch;
						}
						else if(rOtherPatch==maxPatch)
						{
//							feLog("%d shift patch %d into %d\n",
//									pointIndex,rOtherPatch,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPatch=patch;
						}
					}

					for(U32 primitiveOther=0;primitiveOther<primitiveIndex;
							primitiveOther++)
					{
						I32& rOtherPrim=primitivePatch[primitiveOther];
						primitivePatch[primitiveIndex]=lowPatch;
						if(rOtherPrim==patch)
						{
//							feLog("%d merge patch %d into %d\n",
//									primitiveOther,rOtherPrim,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPrim=lowPatch;
						}
						else if(rOtherPrim==maxPatch)
						{
//							feLog("%d shift prim %d into %d\n",
//									primitiveOther,rOtherPrim,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPrim=patch;
						}
					}

					histogram[lowPatch]+=histogram[patch];
					histogram[patch]=histogram[maxPatch];

					FEASSERT(maxPatch);
					maxPatch--;

//					feLog("maxPatch -> %d\n",maxPatch);
				}
			}
		}
		else if(!lowPatch)
		{
			maxPatch++;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=maxPatch;
			}
			lowPatch=maxPatch;
		}
		else if(zeroPatches)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=lowPatch;
			}
		}

//		feLog("primitive %d/%d lowPatch %d/%d\n",
//				primitiveIndex,primitiveCount,lowPatch,maxPatch);

		FEASSERT(lowPatch<=maxPatch);

		FEASSERT(primitiveIndex<primitiveCount);
		primitivePatch[primitiveIndex]=lowPatch;

		FEASSERT(lowPatch<=primitiveCount);
		histogram[lowPatch]+=subCount;

		if(lowPatch)
		{
			FEASSERT(lowPage<=primitiveCount);
			pagePatch[lowPage]=lowPatch;
		}
		if(maxPage<lowPage)
		{
			maxPage=lowPage;
		}
	}

#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPage %d\n",maxPage);
	for(U32 pageIndex=0;pageIndex<=maxPage;pageIndex++)
	{
		FEASSERT(pageIndex<=primitiveCount);
		feLog("page %d uses patch %d\n",pageIndex,pagePatch[pageIndex]);
	}
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPatch %d\n",maxPatch);
	for(U32 patchIndex=0;patchIndex<=maxPatch;patchIndex++)
	{
		FEASSERT(patchIndex<=primitiveCount);
		feLog("patch %d has %d points\n",patchIndex,histogram[patchIndex]);
	}
#endif

	const U32 desiredSize=a_desiredCount? primitiveCount/a_desiredCount: 1;

	Array<U32> patchAtom(maxPatch+1,0);

	I32 maxAtom=1;
	U32 atomSize=0;
	for(U32 patchIndex=1;patchIndex<=maxPatch;patchIndex++)
	{
		if(atomSize>=desiredSize)
		{
			maxAtom++;
			atomSize=0;
		}
		FEASSERT(patchIndex<=maxPatch);
		patchAtom[patchIndex]=maxAtom;

		FEASSERT(patchIndex<=primitiveCount);
		atomSize+=histogram[patchIndex];
#if FE_SAB_DEBUG
		feLog("patch %d atom %d\n",patchIndex,maxAtom);
#endif
	}

	Array<SpannedRange::MultiSpan*> multiSpans(maxAtom+1,NULL);
	I32 spanAtom= -1;
	I32 spanStart= -1;
#if FE_SAB_DEBUG
	feLog("first %d from %d\n",spanAtom,spanStart);
#endif
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		FEASSERT(primitiveIndex<primitiveCount);
		const I32 patchIndex=primitivePatch[primitiveIndex];

		FEASSERT(patchIndex<0 || patchIndex<=maxPatch);
		const I32 atomIndex=patchIndex>=0? patchAtom[patchIndex]: -1;

		FEASSERT(atomIndex<=maxAtom);
		FEASSERT(spanAtom<=maxAtom);
		if(atomIndex!=spanAtom)
		{
			//* complete existing span
			SpannedRange::MultiSpan* pMultiSpan=NULL;
			if(spanAtom>=0)
			{
				FEASSERT(spanAtom<=maxAtom);
				pMultiSpan=multiSpans[spanAtom];
				if(!pMultiSpan)
				{
					pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
							&spSpannedRange->postAtomic();
					multiSpans[spanAtom]=pMultiSpan;
				}
			}
			if(spanStart>=0)
			{
#if FE_SAB_DEBUG
				feLog("complete %d span %d %d\n",
						spanAtom,spanStart,primitiveIndex-1);
#endif
				pMultiSpan->add(spanStart,primitiveIndex-1);
				spanStart= -1;
			}

			//* start another span
			spanAtom=atomIndex;
			if(atomIndex>=0)
			{
				spanStart=primitiveIndex;

#if FE_SAB_DEBUG
				feLog("start %d from %d\n",spanAtom,spanStart);
#endif
			}
		}
	}

	//* complete last span
	SpannedRange::MultiSpan* pMultiSpan=NULL;
	if(spanStart>=0)
	{
		FEASSERT(spanAtom<=maxAtom);
		pMultiSpan=multiSpans[spanAtom];
		if(!pMultiSpan)
		{
			pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
					&spSpannedRange->postAtomic();
			multiSpans[spanAtom]=pMultiSpan;
		}
		pMultiSpan->add(spanStart,primitiveCount-1);
	}

#if FE_SAB_DEBUG
	feLog("brief: %s\n",spSpannedRange->brief().c_str());
	feLog("dump:\n%s\n",spSpannedRange->dump().c_str());
#endif

	return spSpannedRange;
}

} // namespace
